#include "springprox.h"
#include "sprox_desfire.h"

#include <stdio.h>
#include <string.h>

#include "fingertk.h"
#include "sagem_mso_ilv.h"


#define TAG_NONE       0
#define TAG_MIFARE_1K  1
#define TAG_MIFARE_4K  2
#define TAG_DESFIRE    3

BYTE select_tag(BYTE *tag_uid);
void deselect_tag(BYTE tag_type);

WORD swap_word(WORD v)
{
  BYTE t[2];
  WORD w;
  t[0] = (BYTE) (v / 0x0100);
  t[1] = (BYTE) (v % 0x0100);
  memcpy(&w, t, 2);
  return w;
}

/*
 * Read / Write Pro-Active formated tags
 * -------------------------------------
 */
BOOL read_from_tag_mifare(FINGERTK_DATA_ST *data)
{
  BYTE sect;
  BYTE buffer[13][48]; /* 13 sectors of 48 bytes each = 648bytes */
  SWORD rc;

  printf("Reading Mifare tag...\n");
  for (sect = 1; sect <= 13; sect++)
  {
    /* We read with pre-loaded keys (RAM or EEPROM) */
    rc = SPROX_MifStReadSector(NULL, sect, buffer[sect-1], NULL);
    if (rc != MI_OK)
    {
      printf("Failed to read sector %d, cancelled (rc=%d)\n", sect, rc);
      return FALSE;
    }
  }

  /* The 6 first bytes of the received buffer must be our marker */
  if (!memcmp(&buffer[0][0], "PAFNG1", 6))
  {
    /* Assume the tag hold a valid data buffer (more checks shoul'd be done in real-life application) */
    memcpy(data, &buffer[0][6], sizeof(FINGERTK_DATA_ST));

    /* Don't forget to convert WORDs */
    data->finger1_len = swap_word(data->finger1_len);
    data->finger2_len = swap_word(data->finger2_len);
  } else
  {
    /* Try to find a Sagem formated string */
    BYTE tag, idx;
    WORD length, offset;
    BYTE *value;

    offset = 0;
    idx = 0;

    while ((offset < sizeof(buffer)) && ilv_loop(buffer[0], &offset, &tag, &length, &value))
    {
      if ((tag == 0x00) || (tag == 0xFF))
        break;

      switch (tag)
      {
        case 0x32 : /* User ID */
                    if (length > sizeof(data->id))
                      length = sizeof(data->id);
                    memcpy(data->id, value, length);
                    break;

        case 0x20 : /* User Name */
                    if (length > sizeof(data->lastname))
                      length = sizeof(data->lastname);
                    memcpy(data->lastname, value, length);
                    break;

        case 0x30 : /* Finger 1 */
                    if (length <= sizeof(data->finger1_data))
                    {
                      data->finger1_type = ID_PKCOMP; /* Assume it is a PKCOMP minutiae */
                      data->finger1_len = length;
                      memcpy(data->finger1_data, value, length);
                    }
                    break;

        case 0x31 : /* Finger 2 */
                    if (length <= sizeof(data->finger2_data))
                    {
                      data->finger2_type = ID_PKCOMP; /* Assume it is a PKCOMP minutiae */
                      data->finger2_len = length;
                      memcpy(data->finger2_data, value, length);
                    }
                    break;

      }
    }
  }

  /* The 6 first bytes of the received buffer must be our marker */
  if ((data->finger1_len == 0) && (data->finger2_len == 0))
  {
    printf("Tag doesn't hold valid data\n");
    return FALSE;
  }

  printf("Data have been loaded from the Mifare tag\n");
  return TRUE;
}

BOOL read_from_tag_desfire(FINGERTK_DATA_ST *data)
{
  BYTE  buffer[4096];
  BYTE  blank_key[16] = { 0 };
  DWORD size;
  SWORD rc;

  printf("Reading Desfire card...\n");

  /* Try to select our Desfire sample application (AID = 0x016E71) */
	rc = DesFireCard_SelectApplication(0x016E71);
	if (rc != DF_OPERATION_OK)
	{
		printf("Desfire application not found\n");
		goto failed;
	}

  /* Get authenticated on the application with default key (0x00, ..., 0x00) */
  rc = DesFireCard_Authenticate(0, blank_key);
	if (rc != DF_OPERATION_OK)
	{
		printf("Desfire authentication failed (check applications's key 0)\n");
		goto failed;
	}

  /* Read the file holding the data (FID = 1) */
  size = sizeof(buffer);
	rc = DesFireCard_ReadData2(1, 0, 0, buffer, &size);
	if ((rc != DF_OPERATION_OK) || (size == 0))
	{
		printf("Failed to read the Desfire file (rc=%d)\n", rc);
		goto failed;
	}

  /* The information file is a raw copy of the data buffer */
  if (size != sizeof(FINGERTK_DATA_ST))
  {
		printf("Invalid size for the Desfire file\n");
		goto failed;
  }

  /* Copy */
  memcpy(data, buffer, sizeof(FINGERTK_DATA_ST));

  /* Don't forget to convert WORDs */
  data->finger1_len = swap_word(data->finger1_len);
  data->finger2_len = swap_word(data->finger2_len);

  /* Release the card */
  deselect_tag(TAG_DESFIRE);

  /* Be optimistic regarding the actual content of the file... */
  printf("Data have been loaded from the Desfire tag\n");
  return TRUE;

failed:
  /* Don't forget to release the card */
  deselect_tag(TAG_DESFIRE);
  return FALSE;
}

BOOL read_from_tag(FINGERTK_DATA_ST *data)
{
  BYTE tag_type;

  printf("Looking for a card...\n");
  memset(data, 0, sizeof(FINGERTK_DATA_ST));

  tag_type = select_tag(NULL);

  if (tag_type == TAG_DESFIRE)
    return read_from_tag_desfire(data);

  if ((tag_type == TAG_MIFARE_1K) || (tag_type == TAG_MIFARE_4K))
    return read_from_tag_mifare(data);

  return FALSE;
}

void write_to_tag_mifare(FINGERTK_DATA_ST *data)
{
  BYTE sect;
  BYTE buffer[13][48]; /* 13 sectors of 48 bytes each = 648bytes */
  SWORD rc;

  printf("Writing Mifare tag...\n");

  /* The 6 first bytes of the received buffer will be our marker */
  memcpy(&buffer[0][0], "PAFNG1", 6);

  /* Then the data... */
  memcpy(&buffer[0][6], data, sizeof(FINGERTK_DATA_ST));

  for (sect = 1; sect <= 13; sect++)
  {
    /* We read with pre-loaded keys (RAM or EEPROM) */
    rc = SPROX_MifStWriteSector(NULL, sect, buffer[sect-1], NULL);
    if (rc != MI_OK)
    {
      printf("Failed to write sector %d, cancelled\n", sect);
      return;
    }
  }
  
  printf("Data have been saved to the Mifare tag\n");
}

void write_to_tag_desfire(FINGERTK_DATA_ST *data)
{
  FINGERTK_DATA_ST data_copy;
  BYTE  blank_key[16] = { 0 };
  SWORD rc;

  printf("Writing Desfire card...\n");

  /* Try to select our Desfire sample application (AID = 0x016E71) */
	rc = DesFireCard_SelectApplication(0x016E71);
	if (rc != DF_OPERATION_OK)
  {
    /* We must create the application */

    /* Get authenticated on the card with default master key (0x00, ..., 0x00) */
    rc = DesFireCard_Authenticate(0, blank_key);
	  if (rc != DF_OPERATION_OK)
	  {
		  printf("Desfire authentication failed (check card's master key)\n");
		  goto failed;
	  }

    /* Try to create our Desfire sample application (AID = 0x016E71) */
    rc = DesFireCard_CreateApplication(0x016E71, 1, 1);
    if (rc != DF_OPERATION_OK)
    {
		  printf("Failed to create Desfire application\n");
		  goto failed;
    }

    /* Select the newly created application */
	  rc = DesFireCard_SelectApplication(0x016E71);
  }
  
	if (rc != DF_OPERATION_OK)
	{
		printf("Desfire application not found\n");
		goto failed;
	}

  /* Get authenticated on the application with default key (0x00, ..., 0x00) */
  rc = DesFireCard_Authenticate(0, blank_key);
	if (rc != DF_OPERATION_OK)
	{
		printf("Desfire authentication failed (check applications's key 0)\n");
		goto failed;
	}

  /* Delete the (maybe ?) pre-existing file, don't check the returned code */
  DesFireCard_DeleteFile(1);

  /* Create our file (with FID=1) */
  rc = DesFireCard_CreateStdDataFile(1, DF_COMM_MODE_PLAIN, 0xEEEE, sizeof(FINGERTK_DATA_ST));
	if (rc != DF_OPERATION_OK)
	{
		printf("Failed to create Desfire file\n");
		goto failed;
	}

  /* Copy data, and don't forget to convert WORDs */
  data_copy = *data;
  data_copy.finger1_len = swap_word(data_copy.finger1_len);
  data_copy.finger2_len = swap_word(data_copy.finger2_len);

  /* Write the buffer into the file */
	rc = DesFireCard_WriteData2(1, 0, sizeof(FINGERTK_DATA_ST), (BYTE *) &data_copy);
	if (rc != DF_OPERATION_OK)
	{
		printf("Failed to write the Desfire file\n");
		goto failed;
	}

  /* Good */
  printf("Data have been saved to the Desfire tag\n");

failed:
  /* Don't forget to release the card */
  deselect_tag(TAG_DESFIRE);
}

void write_to_tag(FINGERTK_DATA_ST *data)
{
  BYTE tag_type;

  printf("Looking for a card...\n");

  tag_type = select_tag(NULL);

  if (tag_type == TAG_DESFIRE)
  {
    write_to_tag_desfire(data);
  } else
  if ((tag_type == TAG_MIFARE_1K) || (tag_type == TAG_MIFARE_4K))
  {
    write_to_tag_mifare(data);
  }
}



/*
 * Tag manipulation functions
 * --------------------------
 */

BYTE select_tag(BYTE *tag_uid)
{
  BYTE snr[12];
  BYTE snrlen = 12;
  BYTE atq[2];
  BYTE sak[1];
  BYTE ats[32];
	BYTE atslen = 32;
  SWORD rc;

  /* ISO 14443-3 activation, with T=CL compliance check */
  rc = SPROX_TclA_ActivateAny(atq, snr, &snrlen, sak);

  if (tag_uid != NULL)
    memcpy(tag_uid, snr, snrlen);

  if (rc == TCL_NOTAGERR)
  {
    /* Not a T=CL tag, maybe Mifare ? */
    rc = SPROX_MifStSelectAny(snr, atq, sak);
    if (rc != MI_OK)
    {
      printf("Tag activation failed\n");
      return TAG_NONE;
    }
    if (atq[1] == 0x00)
    {
      if (atq[0] == 0x04) /* Mifare 1k */
      {
        printf("Found Mifare 1k tag, UID=%02X%02X%02X%02X\n", snr[3], snr[2], snr[1], snr[0]);
        return TAG_MIFARE_1K;
      }
      if (atq[0] == 0x02) /* Mifare 4k */
      {
        printf("Found Mifare 4k tag, UID=%02X%02X%02X%02X\n", snr[3], snr[2], snr[1], snr[0]);
        return TAG_MIFARE_4K;
      }
    }
    /* Not supported */
    SPROX_TclA_Halt();
    printf("Unsupported tag\n");
    return TAG_NONE;
  } else
  if (rc != MI_OK)
  {
    /* No tag ? */
    printf("No tag found\n");
    return TAG_NONE;
  }

  if ((atq[1] == 0x03) && (atq[0] == 0x44))
  {
    /* This seems to be a Desfire... */
    printf("Found Desfire tag, UID=%02X%02X%02X%02X%02X%02X%02X\n", snr[6], snr[5], snr[4], snr[3], snr[2], snr[1], snr[0]);
    rc = SPROX_TclA_GetAts(0xFF, ats, &atslen);
    if (rc == MI_OK)
    {
      /* Desfire activated */
      return TAG_DESFIRE;
    } else
    {
      printf("Tag activation failed\n");
    }
  } else
  {
    printf("Unsupported tag\n");
  }

  /* Not supported, or no tag activated */
  SPROX_TclA_Halt();
  return TAG_NONE;
}


void deselect_tag(BYTE tag_type)
{
  if (tag_type == TAG_NONE)
    return;
  if (tag_type == TAG_DESFIRE)
    SPROX_TclA_Deselect(0xFF);
  SPROX_TclA_Halt();
}
