using System;
using System.Drawing;
using System.Collections;
using System.Windows.Forms;
using System.Data;
using System.Threading;
using System.IO;

using SpringProxAPI;
using SproxDesFireAPI;

namespace fingertk
{
  public class fMain:System.Windows.Forms.Form
  {
#region Card data

    private const int CARD_UNKNOWN = -1;
    private const int CARD_NONE = 0;
    private const int CARD_MIFARE_1K = 1;
    private const int CARD_MIFARE_4K = 2;
    private const int CARD_MIFARE_UL = 3;
    private const int CARD_DESFIRE = 4;
    private const int CARD_JAYCOS = 5;

    public bool card_is_tcl;
    public bool card_is_type_a;
    public bool card_is_type_b;
    public int card_is_type;

    public byte[] card_iso_a_atq = new byte[2];
    public byte[] card_iso_a_sak = new byte[1];
    public byte[] card_iso_a_snr = new byte[12];
    public byte card_iso_a_snrlen;
    public byte[] card_iso_b_atq = new byte[11];

    public byte[] card_tcl_ats = new byte[32];
    public byte card_tcl_atslen;

    public byte[] card_user_data = new byte[80];

    public byte card_finger_type;
    public byte card_finger_size_1;
    public byte[] card_finger_data_1 = new byte[170];
    public byte card_finger_size_2;
    public byte[] card_finger_data_2 = new byte[170];

    string toHEX(byte[]buffer, int offset, int length)
    {
      int i;
      string r = "";

      for (i = 0; i < length; i++)
          r += String.Format("{0:X02}", buffer[offset + i]);

        return r;
    }

    string toSTR(byte[]buffer, int offset, int length)
    {
      int i;
      string r = "";

      for (i = 0; i < length; i++)
      {
        if (buffer[offset + i] == 0)
          break;
        r += (char) buffer[offset + i];
      }

      return r;
    }

    void fromHEX(byte[]buffer, int offset, int size, string text)
    {
      int i;

      for (i = 0; i < size; i++)
        buffer[offset + i] = 0;

      for (i = 0; i < text.Length; i++)
      {
        if (i >= 2 * size)
          break;

        char c = text[i];

        if (c == '\0')
          break;

        byte b = 0;

        if ((c >= '0') && (c <= '9'))
          b = (byte) (c - '0');
        else if ((c >= 'A') && (c <= 'F'))
          b = (byte) (c + 10 - 'A');
        else if ((c >= 'a') && (c <= 'f'))
          b = (byte) (c + 10 - 'a');

        if ((i % 2) != 0)
        {
          buffer[offset + (i / 2)] |= b;
        } else
        {
          buffer[offset + (i / 2)] = (byte) (b << 4);
        }
      }
    }

    void fromSTR(byte[]buffer, int offset, int size, string text)
    {
      int i;

      for (i = 0; i < size; i++)
        buffer[offset + i] = 0;

      for (i = 0; i < text.Length; i++)
      {
        char c = text[i];

        if (c == '\0')
          break;
        buffer[offset + i] = (byte) c;
      }
    }

#endregion

#region Worker (background thread)

    private Thread workerThread;
    private ManualResetEvent workerEvent;
    private bool workerAllowed;
    private bool workerRunning;
    private int workerNextTask;
    private int workerDoneTask;
    private int workerAfterTask;
    private bool workerResult;
    private bool readerActive;
    private bool readerHasFingerprint;

    private const int WORK_NONE = 0;
    private const int WORK_OPEN_READER = 1;
    private const int WORK_FIND_CARD = 2;
    private const int WORK_READ_CARD = 3;
    private const int WORK_WRITE_CARD = 4;
    private const int WORK_VERIFY = 5;
    private const int WORK_ENROLL = 6;

    private bool ConnectReader()
    {
      short rc;

      if (readerActive)
      {
        rc = SPROX.ControlLed(0, 0, 0);
        if (rc == SPROX.MI_OK)
          return true;
        readerActive = false;
        SPROX.ReaderClose();
      }

      rc = SPROX.ReaderOpen("");
      if (rc == SPROX.MI_OK)
        readerActive = true;

      readerHasFingerprint = false;
      uint features = 0;

      if (SPROX.ReaderGetFeatures(ref features) == SPROX.MI_OK)
      {
        if ((features & 0x00000040) != 0)
          readerHasFingerprint = true;
      }

      return readerActive;
    }

    private bool ReadCardTask()
    {
      bool rc = false;

      for (int i = 0; i < card_user_data.Length; i++)
        card_user_data[i] = 0;

      card_finger_type = 0;
      card_finger_size_1 = 0;
      card_finger_size_2 = 0;

      switch (card_is_type)
      {
        case CARD_MIFARE_1K:
        case CARD_MIFARE_4K:
          rc = ReadCardMifare();
          break;

        case CARD_DESFIRE:
          rc = ReadCardDesfire();
          break;
      }

      /* Power down the antenna to preserve batteries */
      SPROX.ControlRF(0);
      return rc;
    }

    private bool WriteCardTask()
    {
      bool rc = false;

      switch (card_is_type)
      {
        case CARD_MIFARE_1K:
        case CARD_MIFARE_4K:
          rc = WriteCardMifare();
          break;

        case CARD_DESFIRE:
          rc = WriteCardDesfire();
          break;
      }

      /* Power down the antenna to preserve batteries */
      SPROX.ControlRF(0);
      return rc;
    }

    private void workerLoop()
    {
      int task;
      bool rc;

      workerRunning = true;
      while (workerAllowed)
      {
        workerEvent.WaitOne();
        if (!workerAllowed)
          break;

        task = workerNextTask;
        rc = false;

        if (ConnectReader())
        {
          switch (task)
          {
            case WORK_FIND_CARD:
              rc = FindCardTask();
              break;
            case WORK_READ_CARD:
              rc = ReadCardTask();
              break;
            case WORK_WRITE_CARD:
              rc = WriteCardTask();
              break;
            case WORK_VERIFY:
              rc = VerifyTask();
              break;
            case WORK_ENROLL:
              rc = EnrollTask();
              break;
            case WORK_OPEN_READER:
              rc = true;
              break;
          }
        }

        workerResult = rc;
        workerDoneTask = task;
        workerNextTask = 0;
        workerEvent.Reset();

        Invoke(new EventHandler(workerCallback));
      }
      workerRunning = false;
    }

    private void workerCallback(object sender, EventArgs e)
    {
      if (workerDoneTask != 0)
      {
        workerDone(workerDoneTask, workerResult);
        workerDoneTask = 0;
      }
    }

    private void workerDone(int task, bool rc)
    {
      switch (task)
      {
        case WORK_OPEN_READER:
          if (rc)
          {
            if (readerHasFingerprint)
            {
              eStatus.Text = "Reader activated";
              UpdateStatus(true, false, false);
            } else
            {
              eStatus.Text = "No fingerprint device";
            }
          } else
          {
            eStatus.Text = "Reader not available";
          }

          break;

        case WORK_FIND_CARD:
          if (rc)
          {
            switch (card_is_type)
            {
              case CARD_MIFARE_1K:
                eStatus.Text = "Mifare 1k card";
                break;
              case CARD_MIFARE_4K:
                eStatus.Text = "Mifare 4k card";
                break;
              case CARD_DESFIRE:
                eStatus.Text = "Desfire card";
                break;
              default:
                eStatus.Text = "Unsupported card";
                card_is_type = CARD_NONE;
                break;

            }

            if ((workerAfterTask != 0) && (card_is_type != CARD_NONE))
            {
              workerStart(workerAfterTask);
            } else
            {
              UpdateStatus(true);
            }

          } else
          {
            eStatus.Text = "No card found";
            UpdateStatus(true);
          }

          workerAfterTask = 0;
          break;

        case WORK_READ_CARD:
          if (rc)
          {
            eStatus.Text = "Read success";
            eUserID.Text = toHEX(card_user_data, 0, 8);
            eLastName.Text = toSTR(card_user_data, 16, 32);
            eFirstName.Text = toSTR(card_user_data, 48, 32);
            UpdateStatus(true, true, false);
          } else
          {
            eStatus.Text = "Read failed";
            eUserID.Text = "";
            eLastName.Text = "";
            eFirstName.Text = "";
            UpdateStatus(true, false, false);
          }
          break;

        case WORK_WRITE_CARD:
          pImages.Visible = false;
          if (rc)
          {
            eStatus.Text = "Card has been written";
          } else
          {
            eStatus.Text = "Failed to write the card";
          }
          UpdateStatus(true);
          break;

        case WORK_VERIFY:
          pImages.Visible = false;
          UpdateStatus(true);
          if (rc)
          {
            eStatus.Text = "Fingerprint authentication OK";
            lbFingerprintOK.Visible = true;
          } else
          {
            eStatus.Text = "Fingerprint authentication error";
            lbFingerprintKO.Visible = true;
          }
          break;

        case WORK_ENROLL:
          pImages.Visible = false;
          if (rc)
          {
            eStatus.Text = "Fingerprint enrollement OK";
            UpdateStatus(true, true, true);
          } else
          {
            eStatus.Text = "Fingerprint enrollement error";
            UpdateStatus(true, false, false);
          }
          break;

        default:
          eStatus.Text = "???";
          break;
      }
    }

    private void workerStart(int task)
    {
      workerDoneTask = 0;
      workerNextTask = task;
      workerEvent.Set();
    }

#endregion

#region MSO fingerprint verification

    /*
     * ILV decoder
     * -----------
     */

    /* Function to loop inside an ILV list */
    /* ----------------------------------- */
    bool ilv_loop(byte[]buffer, ref ushort offset, ref byte tag,
                  ref ushort length, ref short value_pos)
    {
      ushort l;

      if (buffer == null)
        return false;

      if (buffer[offset] == 0x00)
        return false;


      tag = buffer[offset];
      l = buffer[offset + 2];
      l *= 0x0100;
      l += buffer[offset + 1];
      length = l;

      value_pos = (short) (offset + 3);

      offset += 3;
      offset += l;

      if (l > 1024)
        return false;

      return true;
    }

    /*
     * ILV encoder
     * -----------
     */

    /* Start the building the ILV buffer */
    void ilv_init(byte[]buffer, ref ushort size, byte tag)
    {
      /* Command appears on top of the buffer */
      buffer[0] = tag;

      /* Actual length will be written at last */
      buffer[1] = 0;
      buffer[2] = 0;

      /* OK */
      size = 3;
    }

    /* Done with ILV buffer */
    void ilv_final(byte[]buffer, ref ushort size)
    {
      ushort length;
      ushort temp;

      /* Compute length of the VALUE field */
      if (size < 3)
      {
        /* Oups... */
        size = 0;
        return;
      }

      temp = size;
      temp -= 3;
      length = temp;
      //      length = size - 3;

      /* Write actual length */
      buffer[1] = (byte) (length % 0x0100);
      buffer[2] = (byte) (length / 0x0100);
    }

    /* Add a DWORD value to the ILV buffer */
    void ilv_add_dword(byte[]buffer, ref ushort size, uint value1)
    {
      /* Enqueue the WORD (Little Endian) */
      buffer[size + 0] = (byte) (value1 % 0x00000100);
      buffer[size + 1] = (byte) (value1 / 0x00000100);
      buffer[size + 2] = (byte) (value1 / 0x00010000);
      buffer[size + 3] = (byte) (value1 / 0x01000000);

      /* OK */
      size += 4;
    }

    /* Add a WORD value to the ILV buffer */
    void ilv_add_word(byte[]buffer, ref ushort size, ushort value1)
    {
      /* Enqueue the WORD (Little Endian) */
      buffer[size + 0] = (byte) (value1 % 0x0100);
      buffer[size + 1] = (byte) (value1 / 0x0100);

      /* OK */
      size += 2;
    }

    /* Add a BYTE value to the ILV buffer */
    void ilv_add_byte(byte[]buffer, ref ushort size, byte value1)
    {
      buffer[size + 0] = value1;
      size += 1;
    }

    /* Add a data buffer to the ILV buffer */
    void ilv_add_data(byte[]buffer, ref ushort size, byte[]data,
                      ushort length)
    {
      /* Enqueue the buffer */
      for (ushort i = 0; i < length; i++)
        buffer[size + 0 + i] = data[i];
      size += length;
    }

    private bool VerifyTask()
    {
      byte[]buffer = new byte[1024];
      byte async;
      ushort size;
      ushort length;
      short rc;


      for (int i = 0; i < buffer.Length; i++)
        buffer[i] = 0;
      length = 0;

      /* Initialize ILV */
      ilv_init(buffer, ref length, 0x20); //ILV_VERIFY);

      /* Build ILV */
      ilv_add_word(buffer, ref length, 10); /* Timeout = 10 seconds */
      ilv_add_word(buffer, ref length, 5);  /* Matching threshold = 5 (recommended value) */
      ilv_add_byte(buffer, ref length, 0);  /* Acquisition quality threshold = default (0) */


      /* Add fingers */
      if (card_finger_size_1 > 0)
      {
        ilv_add_byte(buffer, ref length, card_finger_type);
        ilv_add_word(buffer, ref length, card_finger_size_1);
        ilv_add_data(buffer, ref length, card_finger_data_1,
                     card_finger_size_1);
      }
      if (card_finger_size_2 > 0)
      {
        ilv_add_byte(buffer, ref length, card_finger_type);
        ilv_add_word(buffer, ref length, card_finger_size_2);
        ilv_add_data(buffer, ref length, card_finger_data_2,
                     card_finger_size_2);
      }

      /* We want to receive asynchronous events */
      ilv_add_byte(buffer, ref length, 52); //ID_ASYNCHRONOUS_EVENT);
      ilv_add_word(buffer, ref length, 4);
      ilv_add_dword(buffer, ref length, 0x00000001);

      /* Write length in the ILV */
      ilv_final(buffer, ref length);

      /* Send the command to the MSO and process its responses (including asynchronous messages) */
      /* --------------------------------------------------------------------------------------- */
      async = 0;
      for (;;)
      {
        size = (ushort) buffer.Length;

        /* NOTE : the timeout of the exchange must be a little greater than the timeout offered  */
        /*        to the MSO, overwhise the library will return an error code before the MSO has */
        /*        finished is process                                                            */
        if (async != 0)
          rc =
            SPROX.MSO_ExchangeEx(null, 0, buffer, ref size, 12000, ref async);
        else
          rc =
            SPROX.MSO_ExchangeEx(buffer, length, buffer, ref size, 35000,
                                 ref async);

        if (rc != 0)
          return false;         /* Error */

        if (async != 0)
        {
          /* Easy processing of asynchronous message, no error control */
          switch (buffer[4])
          {
            case 0x01:
              VerifyAsyncAction(buffer[7]);
              break;
          }

          /* NOTE : we MUST go back to MSO_Exchange AS SOON AS POSSIBLE because if we were to  */
          /*        miss any message (asynchronous or final result), the reader will report a  */
          /*        protocol error and abort the process                                       */
        } else
        {
          break;                /* Last message -> end of the loop */
        }
      }

      /* 
       * Reply format is :
       *  0      : reply ID
       *  1 -> 2 : reply length
       *  3      : request status
       *  4      : matching result
       */

      /* Check returned reply and status */
      /* ------------------------------- */

      if (buffer[0] != 0x20)    //ILV_VERIFY)
      {
        eStatus.Text = "Wrong MSO reply";
        return false;
      }
      if (buffer[3] != 0x00)
      {
        eStatus.Text = "Wrong MSO status";
        return false;
      }

      /* Check matching result */
      /* --------------------- */

      switch (buffer[4])
      {
        case 1:                //ILVSTS_HIT    : /* Verify -> OK ! */
          eStatus.Text = "Verify -> OK !";
          return true;

        case 2:                //ILVSTS_NO_HIT : /* Verify -> KO ! */
          eStatus.Text = "Verify -> KO !";
          return false;
      }

      eStatus.Text = "Wrong MSO result";
      return false;
    }

    void VerifyAsyncAction(byte code)
    {
      switch (code)
      {
        case 0:
          eStatus.Text = "";
          show_enroll_image(imgVerify);
          show_async_image(imgPlaceFinger);
          break;
        case 1:
          eStatus.Text = "Move up";
          show_async_image(imgMoveUp);
          break;
        case 2:
          eStatus.Text = "Move down";
          show_async_image(imgMoveDown);
          break;
        case 3:
          eStatus.Text = "Move left";
          show_async_image(imgMoveLeft);
          break;
        case 4:
          eStatus.Text = "Move right";
          show_async_image(imgMoveLeft);
          break;
        case 5:
          eStatus.Text = "Press harder";
          show_async_image(imgPressHarder);
          break;
        case 6:
          eStatus.Text = "Move finger";
          break;
        case 7:
          eStatus.Text = "Remove finger";
          show_async_image(imgRemoveFinger);
          break;
        case 8:
          eStatus.Text = "Finger OK";
          show_async_image(imgFingerOK);
          break;
      }

      Application.DoEvents();
    }
    private bool EnrollTask()
    {
      byte[]recv_buffer = new byte[1024];
      byte[]send_buffer = new byte[1024];
      byte async;
      ushort size;
      byte tag;
      ushort offset;
      ushort length;
      short value_pos;
      byte idx;
      short rc;

      card_finger_type = 0;
      card_finger_size_1 = 0;
      card_finger_size_2 = 0;

      /* Build the ILV enroll command               */
      /* (SEE SAGEM DOCUMENTATION FOR EXPLANATIONS) */
      /* ------------------------------------------ */

      length = 0;

      /* Initialize ILV */
      ilv_init(send_buffer, ref length, 0x21);  //ILV_ENROLL);

      /* Build ILV */
      ilv_add_byte(send_buffer, ref length, 0); /* Database identifier = 0 */
      ilv_add_word(send_buffer, ref length, 60);  /* Timeout = 60 seconds */
      ilv_add_byte(send_buffer, ref length, 0); /* Acquisition quality threshold = default (0) */
      ilv_add_byte(send_buffer, ref length, 0); /* Enrollment type = default (0) */
      ilv_add_byte(send_buffer, ref length, 2); /* We want to enroll 2 fingers */
      ilv_add_byte(send_buffer, ref length, 0); /* Save record into local database = NO (0) */
      ilv_add_byte(send_buffer, ref length, 1); /* Export minutiae with its default size = YES (1) */

      /* We want to receive asynchronous events */
      ilv_add_byte(send_buffer, ref length, 52);  //ID_ASYNCHRONOUS_EVENT);
      ilv_add_word(send_buffer, ref length, 4);
      ilv_add_dword(send_buffer, ref length, 0x00000005);
      /* Bit field = 0x00000005 : action messages (move left, press harder...) */
      /* and information messages (finger X/X, capture X/X) will be sent       */

      /* Write length in the ILV */
      ilv_final(send_buffer, ref length);

      /* Send the command to the MSO and process its responses (including asynchronous messages) */
      /* --------------------------------------------------------------------------------------- */
      async = 0;
      for (;;)
      {
        size = 1024;

        /* NOTE : the timeout of the exchange must be a little greater than the timeout offered  */
        /*        to the MSO, overwhise the library will return an error code before the MSO has */
        /*        finished is process                                                            */

        if (async != 0)
          rc =
            SPROX.MSO_ExchangeEx(null, 0, recv_buffer, ref size, 65000,
                                 ref async);
        else
          rc =
            SPROX.MSO_ExchangeEx(send_buffer, length, recv_buffer, ref size,
                                 65000, ref async);

        if (rc != 0)
          return false;

        if (async != 0)
        {
          /* Easy processing of asynchronous message, no error control */
          switch (recv_buffer[4])
          {
            case 0x01:
              EnrollAsyncAction(recv_buffer[7]);
              break;
            case 0x04:
              EnrollAsyncInfo(recv_buffer[7], recv_buffer[8],
                              recv_buffer[9], recv_buffer[10]);
              break;
          }

          /* NOTE : we MUST go back to MSO_Exchange AS SOON AS POSSIBLE because if we were to  */
          /*        miss any message (asynchronous or final result), the reader will report a  */
          /*        protocol error and abort the process                                       */
        } else
        {
          break;                /* Last message -> end of the loop */
        }
      }

      /* 
       * Reply format is :
       *  0      : reply ID
       *  1 -> 2 : reply length
       *  3      : request status
       *  4      : enroll status
       *  5 -> 8 : user database index
       *  9      : first finger
       *  ...    : second finger
       */

      /* Check returned reply and status */
      /* ------------------------------- */

      if (recv_buffer[0] != 0x21) //ILV_ENROLL)
      {
        eStatus.Text = "Wrong MSO reply";
        return false;
      }
      if (recv_buffer[3] != 0x00)
      {
        eStatus.Text = "Wrong MSO status";
        return false;
      }
      if (recv_buffer[4] != 0x00)
      {
        eStatus.Text = "Wrong MSO result";
        return false;
      }

      /* Decode returned data */
      /* -------------------- */

      offset = 9;
      idx = 0;
      tag = 0;

      value_pos = 0;
      while ((offset <
              size) & (ilv_loop(recv_buffer, ref offset, ref tag, ref length,
                                ref value_pos)))
      {
        if ((tag != 2)          //ID_PKCOMP)  /* PKCOMP minutiae */
            && (tag != 3))      //ID_PKMAT))  /* PKMAT minutiae */
        {
          /* Skip */
          continue;
        }

        if (idx < 2)
        {
          /* Finger 1 or 2 */
          if (length > 256)
          {
            continue;
          }
          if (idx == 0)
          {
            card_finger_type = tag;
            card_finger_size_1 = (byte) length;
            for (int i = 0; i < length; i++)
              card_finger_data_1[i] = recv_buffer[value_pos + i];
          } else if (tag == card_finger_type)
          {
            card_finger_size_2 = (byte) length;
            for (int i = 0; i < length; i++)
              card_finger_data_2[i] = recv_buffer[value_pos + i];
          }
          idx++;
        } else
        {
          /* Silently ignore the finger */
        }
      }

      return true;
    }

    void EnrollAsyncAction(byte code)
    {
      switch (code)
      {
        case 0:
          eStatus.Text = "";
          show_async_image(imgPlaceFinger);
          break;
        case 1:
          eStatus.Text = "Move up";
          show_async_image(imgMoveUp);
          break;
        case 2:
          eStatus.Text = "Move down";
          show_async_image(imgMoveDown);
          break;
        case 3:
          eStatus.Text = "Move left";
          show_async_image(imgMoveLeft);
          break;
        case 4:
          eStatus.Text = "Move right";
          show_async_image(imgMoveLeft);
          break;
        case 5:
          eStatus.Text = "Press harder";
          show_async_image(imgPressHarder);
          break;
        case 6:
          eStatus.Text = "Move finger";
          show_async_image(imgPlaceFinger);
          break;
        case 7:
          if (!imgFingerOK.Visible)
            show_async_image(imgRemoveFinger);
          break;
        case 8:
          eStatus.Text = "Finger OK";
          show_async_image(imgFingerOK);
          break;
      }

      Application.DoEvents();
    }

    void EnrollAsyncInfo(byte finger_no, byte finger_count, byte capture_no,
                         byte capture_count)
    {
      eStatus.Text =
        String.Format("Finger {0}/{1}, capture {2}/{3}", finger_no,
                      finger_count, capture_no, capture_count);
      if (finger_no == 1)
      {
        if (capture_no == 1)
          show_enroll_image(imgEnrollLeft1);
        else if (capture_no == 2)
          show_enroll_image(imgEnrollLeft2);
        else if (capture_no == 3)
          show_enroll_image(imgEnrollLeft3);
        else
          show_enroll_image(null);
      } else if (finger_no == 2)
      {
        if (capture_no == 1)
          show_enroll_image(imgEnrollRight1);
        else if (capture_no == 2)
          show_enroll_image(imgEnrollRight2);
        else if (capture_no == 3)
          show_enroll_image(imgEnrollRight3);
        else
          show_enroll_image(null);
      } else
        show_enroll_image(null);

      Application.DoEvents();
    }


#endregion

#region Mifare read & write

    public byte[] MIFARE_HEADER =
      new byte[16] { 0x00, 0x00, 0x50, 0x72, 0x6F, 0x2D, 0x41, 0x63, 0x74,
      0x69, 0x76, 0x65, 0x00, 0x11, 0x22, 0x33
    };
    public byte[] MIFARE_KEY_A =
      new byte[6] { 0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5 };
    public byte[] MIFARE_KEY_B =
      new byte[6] { 0x50, 0x41, 0x00, 0x11, 0x22, 0x33 };

    private bool ReadCardMifare()
    {
      byte[]buffer = new byte[528];
      int i;
      byte sector;
      short rc;

      /* Read the part common to 1k and 4k */
      for (sector = 0; sector < 11; sector++)
      {
        byte[]t = new byte[48];
        rc = SPROX.MifStReadSector(null, sector, t, MIFARE_KEY_A);
        if (rc != SPROX.MI_OK)
          return false;
        for (i = 0; i < 48; i++)
          buffer[48 * sector + i] = t[i];

        /* Check that we retrieve our header in block 1 */
        if (sector == 0)
          for (i = 0; i < 16; i++)
            if (t[16 + i] != MIFARE_HEADER[i])
              return false;
      }

      /* Retrieve user data from offset 48 */
      for (i = 0; i < 80; i++)
        card_user_data[i] = buffer[48 + i];

      card_finger_size_1 = card_user_data[10];
      card_finger_size_2 = card_user_data[11];
      card_finger_type = card_user_data[12];

      /* Retrieve fingerprints */
      if ((card_finger_size_1 > 0) || (card_finger_size_2 > 0))
      {
        for (i = 0; i < 170; i++)
          card_finger_data_1[i] = buffer[128 + i];
        for (i = 0; i < 170; i++)
          card_finger_data_2[i] = buffer[128 + i + 170];
      }

      return true;
    }

    private bool WriteCardMifare()
    {
      byte[]buffer = new byte[528];
      int i;
      byte sector;
      short rc;

      for (i = 0; i < 528; i++)
        buffer[i] = 0;

      /* Build the header */
      for (i = 0; i < 16; i++)
        buffer[16 + i] = MIFARE_HEADER[i];

      /* Retrieve the information block */
      for (i = 0; i < 80; i++)
        buffer[48 + i] = card_user_data[i];

      /* Retrieve fingerprints */
      if ((card_finger_size_1 > 0) || (card_finger_size_2 > 0))
      {
        for (i = 0; i < 170; i++)
          buffer[128 + i] = card_finger_data_1[i];
        for (i = 0; i < 170; i++)
          buffer[298 + i] = card_finger_data_2[i];
      }

      /* Write the part common to 1k and 4k */
      for (sector = 0; sector < 11; sector++)
      {
        byte[]t = new byte[48];
        for (i = 0; i < 48; i++)
          t[i] = buffer[48 * sector + i];

        /* Try to write with the expected key */
        rc = SPROX.MifStWriteSector(null, sector, t, MIFARE_KEY_B);
        if (rc != SPROX.MI_OK)
        {
          /* Try to format the sector */
          rc =
            SPROX.MifStUpdateAccessBlock(null, sector, null, MIFARE_KEY_A,
                                         MIFARE_KEY_B, SPROX.ACC_BLOCK_VALUE,
                                         SPROX.ACC_BLOCK_VALUE,
                                         SPROX.ACC_BLOCK_VALUE,
                                         SPROX.ACC_AUTH_NORMAL);
          if (rc != SPROX.MI_OK)
            return false;
          /* Retry to write with the expected key */
          rc = SPROX.MifStWriteSector(null, sector, t, MIFARE_KEY_B);
          if (rc != SPROX.MI_OK)
            return false;
        }
      }

      return true;
    }

#endregion

#region Desfire read & write

    public const uint DESFIRE_APPLICATION_ID = 0x00112233;
    public const byte DESFIRE_IDENT_FILE_ID = 0;
    public const byte DESFIRE_PICTURE_FILE_ID = 1;
    public const byte DESFIRE_FINGER_FILE_ID = 2;
    public const byte DESFIRE_BALANCE_FILE_ID = 3;
    public const byte DESFIRE_HISTORY_FILE_ID = 4;
    public byte[] DESFIRE_BLANK_KEY =
      new byte[16] { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    };
    public byte[] DESFIRE_KEY_0 =
      new byte[16] { 0xB0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x11, 0x22, 0x33
    };
    public byte[] DESFIRE_KEY_1 =
      new byte[16] { 0xA0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x11, 0x22, 0x33
    };

    private bool ReadCardDesfire()
    {
      uint length = 0;
      int i;

      if (SproxDesFire.SelectApplication(DESFIRE_APPLICATION_ID) !=
          SPROX.MI_OK)
        return false;

      if (SproxDesFire.Authenticate(1, DESFIRE_KEY_1) != SPROX.MI_OK)
        return false;

      if (SproxDesFire.
          ReadData2(DESFIRE_IDENT_FILE_ID, 0, 0, card_user_data,
                    ref length) != SPROX.MI_OK)
        return false;

      card_finger_size_1 = card_user_data[10];
      card_finger_size_2 = card_user_data[11];
      card_finger_type = card_user_data[12];

      if ((card_finger_size_1 > 0) || (card_finger_size_2 > 0))
      {
        byte[]buffer = new byte[340];

        if (SproxDesFire.
            ReadData2(DESFIRE_FINGER_FILE_ID, 0, 0, buffer,
                      ref length) != SPROX.MI_OK)
          return false;
        for (i = 0; i < 170; i++)
          card_finger_data_1[i] = buffer[i];
        for (i = 0; i < 170; i++)
          card_finger_data_2[i] = buffer[i + 170];
      }


      return true;
    }

    private bool FormatCardDesfire()
    {
      short rc;

      /* Select root application in the Desfire card */
      rc = SproxDesFire.SelectApplication(0);
      if (rc != SPROX.MI_OK)
        return false;

      /* Perform 3-DES authentication with card's master key (must be blank !) */
      rc = SproxDesFire.Authenticate(0, DESFIRE_BLANK_KEY);
      if (rc != SPROX.MI_OK)
        return false;

      /* Create our application. Application has 2 keys. We don't care of key_settings here. */
      rc = SproxDesFire.CreateApplication(DESFIRE_APPLICATION_ID, 0xE3, 2);
      if (rc != SPROX.MI_OK)
        return false;

      /* Select the newly created application */
      rc = SproxDesFire.SelectApplication(DESFIRE_APPLICATION_ID);
      if (rc != SPROX.MI_OK)
        return false;

      /* Get authenticated with blank key 0 */
      rc = SproxDesFire.Authenticate(0, DESFIRE_BLANK_KEY);
      if (rc != SPROX.MI_OK)
        return false;

      /* Create the files */
      rc =
        SproxDesFire.CreateStdDataFile(DESFIRE_IDENT_FILE_ID, 0, 0x1F00, 80);
      if (rc != SPROX.MI_OK)
        return false;
      rc =
        SproxDesFire.CreateStdDataFile(DESFIRE_PICTURE_FILE_ID, 0, 0x1F00,
                                       3072);
      if (rc != SPROX.MI_OK)
        return false;
      rc =
        SproxDesFire.CreateStdDataFile(DESFIRE_FINGER_FILE_ID, 3, 0x1F00,
                                       340);
      if (rc != SPROX.MI_OK)
        return false;
      rc =
        SproxDesFire.CreateValueFile(DESFIRE_BALANCE_FILE_ID, 2, 0x1F00, 0,
                                     100000, 0, 0);
      if (rc != SPROX.MI_OK)
        return false;
      rc =
        SproxDesFire.CreateCyclicRecordFile(DESFIRE_HISTORY_FILE_ID, 0,
                                            0xEEEE, 8, 4);
      if (rc != SPROX.MI_OK)
        return false;

      rc = SproxDesFire.ChangeKey(0, DESFIRE_KEY_0, null);
      if (rc != SPROX.MI_OK)
        return false;
      rc = SproxDesFire.Authenticate(1, DESFIRE_BLANK_KEY);
      if (rc != SPROX.MI_OK)
        return false;
      rc = SproxDesFire.ChangeKey(1, DESFIRE_KEY_1, null);
      if (rc != SPROX.MI_OK)
        return false;

      return true;
    }

    private bool WriteCardDesfire()
    {
      byte[]buffer = new byte[4096];
      int i;

      /* Select the application, or create it if not found */
      if (SproxDesFire.SelectApplication(DESFIRE_APPLICATION_ID) !=
          SPROX.MI_OK)
        if (!FormatCardDesfire())
          return false;

      /* Get authenticated with key 0 for writing */
      if (SproxDesFire.Authenticate(0, DESFIRE_KEY_0) != SPROX.MI_OK)
        return false;

      /* Write the information block */
      if (SproxDesFire.
          WriteData2(DESFIRE_IDENT_FILE_ID, 0, 80,
                     card_user_data) != SPROX.MI_OK)
        return false;

      /* Write the fingerprints block */
      if ((card_finger_size_1 > 0) || (card_finger_size_2 > 0))
      {
        for (i = 0; i < 170; i++)
          buffer[i] = card_finger_data_1[i];
        for (i = 0; i < 170; i++)
          buffer[i + 170] = card_finger_data_2[i];

        if (SproxDesFire.WriteData2(DESFIRE_FINGER_FILE_ID, 0, 340, buffer) !=
            SPROX.MI_OK)
          return false;
      }

      return true;
    }

#endregion

#region ISO/IEC 14443 card detection and selection

    private bool FindCardTask()
    {
      short rc;

      if (!readerActive)
        return false;

      /* We must halt the card first (in case we were already connected) */
      if (card_is_tcl)
      {
        SPROX.Tcl_Deselect(0xFF);
      } else if (card_is_type_a)
      {
        SPROX.A_Halt();
      } else if (card_is_type_b)
      {
        SPROX.B_Halt(card_iso_b_atq);
      }

      card_is_type_a = false;
      card_is_type_b = false;
      card_is_tcl = false;

      SPROX.ControlRF(1);

      /* At first try to find an ISO-14443-A card */
      SPROX.SetConfig(SPROX.CFG_MODE_ISO_14443_A);

      card_iso_a_snrlen = 12;
      rc =
        SPROX.A_SelectAny(card_iso_a_atq, card_iso_a_snr,
                          ref card_iso_a_snrlen, card_iso_a_sak);

      if (rc == SPROX.MI_OK)
      {
        /* A card has been found */
        card_is_type = CARD_UNKNOWN;

        if ((card_iso_a_sak[0] & 0x20) != 0)
        {
          /* Card support T=CL, try to activate it */
          card_tcl_atslen = 32;
          rc = SPROX.TclA_GetAts(0xFF, card_tcl_ats, ref card_tcl_atslen);
          if (rc == SPROX.MI_OK)
          {
            /* Found a type A T=CL card */
            card_is_type_a = true;
            card_is_tcl = true;

            if ((card_iso_a_atq[1] == 0x03) && (card_iso_a_atq[0] == 0x44))
            {
              card_is_type = CARD_DESFIRE;
            }
            return true;
          }
        } else
        {
          /* No T=CL support, but type A card found anyway */
          card_is_type_a = true;

          if ((card_iso_a_atq[1] == 0x00) && (card_iso_a_atq[0] == 0x04))
          {
            card_is_type = CARD_MIFARE_1K;
          } else
            if ((card_iso_a_atq[1] == 0x00) && (card_iso_a_atq[0] == 0x02))
          {
            card_is_type = CARD_MIFARE_4K;
          } else
            if ((card_iso_a_atq[1] == 0x03) && (card_iso_a_atq[0] == 0x40))
          {
            card_is_type = CARD_MIFARE_UL;
          }

          return true;
        }

        /* Type A card found but with error */
        goto failed;
      }

      /* Nox try to find an ISO-14443-B card */
      SPROX.SetConfig(SPROX.CFG_MODE_ISO_14443_B);

      rc = SPROX.B_SelectAny(0x00, card_iso_b_atq);

      if (rc == SPROX.MI_OK)
      {
        /* A card has been found */
        card_is_type = CARD_UNKNOWN;

        if ((card_iso_b_atq[10] & 0x01) != 0)
        {
          /* Card support T=CL, try to activate it */
          card_tcl_atslen = 32;
          rc =
            SPROX.TclB_Attrib(card_iso_b_atq, 0xFF, card_tcl_ats,
                              ref card_tcl_atslen);
          if (rc == SPROX.MI_OK)
          {
            /* Found a type B T=CL card */
            card_is_type_b = true;
            card_is_tcl = true;
            return true;
          }
        } else
        {
          /* No T=CL support, but type Bcard found anyway */
          card_is_type_b = true;
          return true;
        }

        /* Type B card found but with error */
        goto failed;
      }

    failed:
      /* Power down the antenna to preserve batteries */
      SPROX.ControlRF(0);
      return false;
    }

#endregion

#region Windows Form Design

    private System.Windows.Forms.Button btnEnroll;
    private System.Windows.Forms.Button btnVerify;
    private System.Windows.Forms.Button btnRead;
    private System.Windows.Forms.Button btnWrite;
    private System.Windows.Forms.PictureBox imgEnrollLeft1;
    private System.Windows.Forms.PictureBox imgEnrollLeft2;
    private System.Windows.Forms.PictureBox imgEnrollLeft3;
    private System.Windows.Forms.PictureBox imgEnrollRight1;
    private System.Windows.Forms.PictureBox imgEnrollRight2;
    private System.Windows.Forms.PictureBox imgEnrollRight3;
    private System.Windows.Forms.PictureBox imgFingerOK;
    private System.Windows.Forms.PictureBox imgMoveRight;
    private System.Windows.Forms.PictureBox imgMoveLeft;
    private System.Windows.Forms.PictureBox imgMoveDown;
    private System.Windows.Forms.PictureBox imgMoveUp;
    private System.Windows.Forms.PictureBox imgPlaceFinger;
    private System.Windows.Forms.PictureBox imgRemoveFinger;
    private System.Windows.Forms.PictureBox imgPressHarder;
    private System.Windows.Forms.PictureBox imgVerify;
    private System.Windows.Forms.Panel pImages;
    private System.Windows.Forms.TextBox eFirstName;
    private System.Windows.Forms.Label lbName;
    private System.Windows.Forms.TextBox eLastName;
    private System.Windows.Forms.TextBox eUserID;
    private System.Windows.Forms.Label lbWeb;
    private System.Windows.Forms.Label lbDesc;
    private System.Windows.Forms.TextBox eStatus;
    private System.Windows.Forms.Label lbFingerprintKO;
    private System.Windows.Forms.Label lbFingerprintOK;
    private System.Windows.Forms.Button btnLoad;
    private System.Windows.Forms.Button btnSave;
    private System.Windows.Forms.OpenFileDialog openFileDialog;
    private System.Windows.Forms.SaveFileDialog saveFileDialog;
    private System.Windows.Forms.Label lbUserID;
    private System.Windows.Forms.Label lbAppId;
    private System.Windows.Forms.Label lbCard;
    private System.Windows.Forms.Label lbFile;
    private System.Windows.Forms.Label lbFinger;

    private void InitializeComponent()
    {
      System.ComponentModel.ComponentResourceManager resources = new System.ComponentModel.ComponentResourceManager(typeof(fMain));
      this.eUserID = new System.Windows.Forms.TextBox();
      this.eLastName = new System.Windows.Forms.TextBox();
      this.eFirstName = new System.Windows.Forms.TextBox();
      this.btnRead = new System.Windows.Forms.Button();
      this.btnVerify = new System.Windows.Forms.Button();
      this.btnEnroll = new System.Windows.Forms.Button();
      this.btnWrite = new System.Windows.Forms.Button();
      this.pImages = new System.Windows.Forms.Panel();
      this.imgVerify = new System.Windows.Forms.PictureBox();
      this.imgEnrollRight3 = new System.Windows.Forms.PictureBox();
      this.imgEnrollRight2 = new System.Windows.Forms.PictureBox();
      this.imgEnrollRight1 = new System.Windows.Forms.PictureBox();
      this.imgEnrollLeft3 = new System.Windows.Forms.PictureBox();
      this.imgEnrollLeft2 = new System.Windows.Forms.PictureBox();
      this.imgEnrollLeft1 = new System.Windows.Forms.PictureBox();
      this.imgPlaceFinger = new System.Windows.Forms.PictureBox();
      this.imgMoveUp = new System.Windows.Forms.PictureBox();
      this.imgMoveDown = new System.Windows.Forms.PictureBox();
      this.imgMoveLeft = new System.Windows.Forms.PictureBox();
      this.imgMoveRight = new System.Windows.Forms.PictureBox();
      this.imgFingerOK = new System.Windows.Forms.PictureBox();
      this.imgPressHarder = new System.Windows.Forms.PictureBox();
      this.imgRemoveFinger = new System.Windows.Forms.PictureBox();
      this.lbName = new System.Windows.Forms.Label();
      this.lbUserID = new System.Windows.Forms.Label();
      this.lbAppId = new System.Windows.Forms.Label();
      this.lbWeb = new System.Windows.Forms.Label();
      this.lbDesc = new System.Windows.Forms.Label();
      this.eStatus = new System.Windows.Forms.TextBox();
      this.lbFingerprintKO = new System.Windows.Forms.Label();
      this.lbFingerprintOK = new System.Windows.Forms.Label();
      this.btnLoad = new System.Windows.Forms.Button();
      this.btnSave = new System.Windows.Forms.Button();
      this.openFileDialog = new System.Windows.Forms.OpenFileDialog();
      this.saveFileDialog = new System.Windows.Forms.SaveFileDialog();
      this.lbCard = new System.Windows.Forms.Label();
      this.lbFinger = new System.Windows.Forms.Label();
      this.lbFile = new System.Windows.Forms.Label();
      // 
      // eUserID
      // 
      this.eUserID.BackColor = System.Drawing.SystemColors.Control;
      this.eUserID.Location = new System.Drawing.Point(0, 80);
      this.eUserID.MaxLength = 16;
      this.eUserID.ReadOnly = true;
      this.eUserID.Size = new System.Drawing.Size(128, 21);
      // 
      // eLastName
      // 
      this.eLastName.BackColor = System.Drawing.SystemColors.Control;
      this.eLastName.Font = new System.Drawing.Font("Microsoft Sans Serif", 9F, System.Drawing.FontStyle.Bold);
      this.eLastName.Location = new System.Drawing.Point(0, 16);
      this.eLastName.MaxLength = 32;
      this.eLastName.ReadOnly = true;
      this.eLastName.Size = new System.Drawing.Size(240, 20);
      // 
      // eFirstName
      // 
      this.eFirstName.BackColor = System.Drawing.SystemColors.Control;
      this.eFirstName.Font = new System.Drawing.Font("Microsoft Sans Serif", 9F, System.Drawing.FontStyle.Bold);
      this.eFirstName.Location = new System.Drawing.Point(0, 36);
      this.eFirstName.MaxLength = 32;
      this.eFirstName.ReadOnly = true;
      this.eFirstName.Size = new System.Drawing.Size(240, 20);
      // 
      // btnRead
      // 
      this.btnRead.Enabled = false;
      this.btnRead.Location = new System.Drawing.Point(0, 224);
      this.btnRead.Size = new System.Drawing.Size(80, 20);
      this.btnRead.Text = "Read";
      this.btnRead.Click += new System.EventHandler(this.btnRead_Click);
      // 
      // btnVerify
      // 
      this.btnVerify.Enabled = false;
      this.btnVerify.Location = new System.Drawing.Point(160, 224);
      this.btnVerify.Size = new System.Drawing.Size(80, 20);
      this.btnVerify.Text = "Verify";
      this.btnVerify.Click += new System.EventHandler(this.btnVerify_Click);
      // 
      // btnEnroll
      // 
      this.btnEnroll.Enabled = false;
      this.btnEnroll.Location = new System.Drawing.Point(184, 200);
      this.btnEnroll.Size = new System.Drawing.Size(56, 20);
      this.btnEnroll.Text = "Enroll";
      this.btnEnroll.Click += new System.EventHandler(this.btnEnroll_Click);
      // 
      // btnWrite
      // 
      this.btnWrite.Enabled = false;
      this.btnWrite.Location = new System.Drawing.Point(0, 200);
      this.btnWrite.Size = new System.Drawing.Size(56, 20);
      this.btnWrite.Text = "Write";
      this.btnWrite.Click += new System.EventHandler(this.btnWrite_Click);
      // 
      // pImages
      // 
      this.pImages.BackColor = System.Drawing.Color.Blue;
      this.pImages.Controls.Add(this.imgVerify);
      this.pImages.Controls.Add(this.imgEnrollRight3);
      this.pImages.Controls.Add(this.imgEnrollRight2);
      this.pImages.Controls.Add(this.imgEnrollRight1);
      this.pImages.Controls.Add(this.imgEnrollLeft3);
      this.pImages.Controls.Add(this.imgEnrollLeft2);
      this.pImages.Controls.Add(this.imgEnrollLeft1);
      this.pImages.Controls.Add(this.imgPlaceFinger);
      this.pImages.Controls.Add(this.imgMoveUp);
      this.pImages.Controls.Add(this.imgMoveDown);
      this.pImages.Controls.Add(this.imgMoveLeft);
      this.pImages.Controls.Add(this.imgMoveRight);
      this.pImages.Controls.Add(this.imgFingerOK);
      this.pImages.Controls.Add(this.imgPressHarder);
      this.pImages.Controls.Add(this.imgRemoveFinger);
      this.pImages.Location = new System.Drawing.Point(24, 24);
      this.pImages.Size = new System.Drawing.Size(188, 164);
      this.pImages.Visible = false;
      // 
      // imgVerify
      // 
      this.imgVerify.Image = ((System.Drawing.Image)(resources.GetObject("imgVerify.Image")));
      this.imgVerify.Location = new System.Drawing.Point(4, 4);
      this.imgVerify.Size = new System.Drawing.Size(180, 80);
      this.imgVerify.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgVerify.Visible = false;
      // 
      // imgEnrollRight3
      // 
      this.imgEnrollRight3.Image = ((System.Drawing.Image)(resources.GetObject("imgEnrollRight3.Image")));
      this.imgEnrollRight3.Location = new System.Drawing.Point(4, 4);
      this.imgEnrollRight3.Size = new System.Drawing.Size(180, 80);
      this.imgEnrollRight3.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgEnrollRight3.Visible = false;
      // 
      // imgEnrollRight2
      // 
      this.imgEnrollRight2.Image = ((System.Drawing.Image)(resources.GetObject("imgEnrollRight2.Image")));
      this.imgEnrollRight2.Location = new System.Drawing.Point(4, 4);
      this.imgEnrollRight2.Size = new System.Drawing.Size(180, 80);
      this.imgEnrollRight2.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgEnrollRight2.Visible = false;
      // 
      // imgEnrollRight1
      // 
      this.imgEnrollRight1.Image = ((System.Drawing.Image)(resources.GetObject("imgEnrollRight1.Image")));
      this.imgEnrollRight1.Location = new System.Drawing.Point(4, 4);
      this.imgEnrollRight1.Size = new System.Drawing.Size(180, 80);
      this.imgEnrollRight1.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgEnrollRight1.Visible = false;
      // 
      // imgEnrollLeft3
      // 
      this.imgEnrollLeft3.Image = ((System.Drawing.Image)(resources.GetObject("imgEnrollLeft3.Image")));
      this.imgEnrollLeft3.Location = new System.Drawing.Point(4, 4);
      this.imgEnrollLeft3.Size = new System.Drawing.Size(180, 80);
      this.imgEnrollLeft3.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgEnrollLeft3.Visible = false;
      // 
      // imgEnrollLeft2
      // 
      this.imgEnrollLeft2.Image = ((System.Drawing.Image)(resources.GetObject("imgEnrollLeft2.Image")));
      this.imgEnrollLeft2.Location = new System.Drawing.Point(4, 4);
      this.imgEnrollLeft2.Size = new System.Drawing.Size(180, 80);
      this.imgEnrollLeft2.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgEnrollLeft2.Visible = false;
      // 
      // imgEnrollLeft1
      // 
      this.imgEnrollLeft1.Image = ((System.Drawing.Image)(resources.GetObject("imgEnrollLeft1.Image")));
      this.imgEnrollLeft1.Location = new System.Drawing.Point(4, 4);
      this.imgEnrollLeft1.Size = new System.Drawing.Size(180, 80);
      this.imgEnrollLeft1.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgEnrollLeft1.Visible = false;
      // 
      // imgPlaceFinger
      // 
      this.imgPlaceFinger.Image = ((System.Drawing.Image)(resources.GetObject("imgPlaceFinger.Image")));
      this.imgPlaceFinger.Location = new System.Drawing.Point(4, 80);
      this.imgPlaceFinger.Size = new System.Drawing.Size(180, 80);
      this.imgPlaceFinger.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgPlaceFinger.Visible = false;
      // 
      // imgMoveUp
      // 
      this.imgMoveUp.Image = ((System.Drawing.Image)(resources.GetObject("imgMoveUp.Image")));
      this.imgMoveUp.Location = new System.Drawing.Point(4, 80);
      this.imgMoveUp.Size = new System.Drawing.Size(180, 80);
      this.imgMoveUp.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgMoveUp.Visible = false;
      // 
      // imgMoveDown
      // 
      this.imgMoveDown.Image = ((System.Drawing.Image)(resources.GetObject("imgMoveDown.Image")));
      this.imgMoveDown.Location = new System.Drawing.Point(4, 80);
      this.imgMoveDown.Size = new System.Drawing.Size(180, 80);
      this.imgMoveDown.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgMoveDown.Visible = false;
      // 
      // imgMoveLeft
      // 
      this.imgMoveLeft.Image = ((System.Drawing.Image)(resources.GetObject("imgMoveLeft.Image")));
      this.imgMoveLeft.Location = new System.Drawing.Point(4, 80);
      this.imgMoveLeft.Size = new System.Drawing.Size(180, 80);
      this.imgMoveLeft.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgMoveLeft.Visible = false;
      // 
      // imgMoveRight
      // 
      this.imgMoveRight.Image = ((System.Drawing.Image)(resources.GetObject("imgMoveRight.Image")));
      this.imgMoveRight.Location = new System.Drawing.Point(4, 80);
      this.imgMoveRight.Size = new System.Drawing.Size(180, 80);
      this.imgMoveRight.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgMoveRight.Visible = false;
      // 
      // imgFingerOK
      // 
      this.imgFingerOK.Image = ((System.Drawing.Image)(resources.GetObject("imgFingerOK.Image")));
      this.imgFingerOK.Location = new System.Drawing.Point(4, 80);
      this.imgFingerOK.Size = new System.Drawing.Size(180, 80);
      this.imgFingerOK.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgFingerOK.Visible = false;
      // 
      // imgPressHarder
      // 
      this.imgPressHarder.Image = ((System.Drawing.Image)(resources.GetObject("imgPressHarder.Image")));
      this.imgPressHarder.Location = new System.Drawing.Point(4, 80);
      this.imgPressHarder.Size = new System.Drawing.Size(180, 80);
      this.imgPressHarder.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgPressHarder.Visible = false;
      // 
      // imgRemoveFinger
      // 
      this.imgRemoveFinger.Image = ((System.Drawing.Image)(resources.GetObject("imgRemoveFinger.Image")));
      this.imgRemoveFinger.Location = new System.Drawing.Point(4, 80);
      this.imgRemoveFinger.Size = new System.Drawing.Size(180, 80);
      this.imgRemoveFinger.SizeMode = System.Windows.Forms.PictureBoxSizeMode.CenterImage;
      this.imgRemoveFinger.Visible = false;
      // 
      // lbName
      // 
      this.lbName.Location = new System.Drawing.Point(0, 0);
      this.lbName.Size = new System.Drawing.Size(100, 16);
      this.lbName.Text = "Name :";
      // 
      // lbUserID
      // 
      this.lbUserID.Location = new System.Drawing.Point(0, 64);
      this.lbUserID.Size = new System.Drawing.Size(80, 16);
      this.lbUserID.Text = "User ID :";
      // 
      // lbAppId
      // 
      this.lbAppId.ForeColor = System.Drawing.Color.FromArgb(((int)(((byte)(0)))), ((int)(((byte)(0)))), ((int)(((byte)(192)))));
      this.lbAppId.Location = new System.Drawing.Point(0, 128);
      this.lbAppId.Size = new System.Drawing.Size(232, 16);
      this.lbAppId.Text = "FingerTK (C# .NET) 19/11/07";
      this.lbAppId.TextAlign = System.Drawing.ContentAlignment.TopCenter;
      // 
      // lbWeb
      // 
      this.lbWeb.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Bold);
      this.lbWeb.ForeColor = System.Drawing.Color.FromArgb(((int)(((byte)(0)))), ((int)(((byte)(0)))), ((int)(((byte)(192)))));
      this.lbWeb.Location = new System.Drawing.Point(0, 144);
      this.lbWeb.Size = new System.Drawing.Size(232, 16);
      this.lbWeb.Text = "www.pro-active.fr";
      this.lbWeb.TextAlign = System.Drawing.ContentAlignment.TopCenter;
      // 
      // lbDesc
      // 
      this.lbDesc.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Bold);
      this.lbDesc.ForeColor = System.Drawing.Color.FromArgb(((int)(((byte)(0)))), ((int)(((byte)(0)))), ((int)(((byte)(192)))));
      this.lbDesc.Location = new System.Drawing.Point(0, 112);
      this.lbDesc.Size = new System.Drawing.Size(232, 16);
      this.lbDesc.Text = "SpringBio SDK sample";
      this.lbDesc.TextAlign = System.Drawing.ContentAlignment.TopCenter;
      // 
      // eStatus
      // 
      this.eStatus.BackColor = System.Drawing.SystemColors.Control;
      this.eStatus.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, ((System.Drawing.FontStyle)((System.Drawing.FontStyle.Bold | System.Drawing.FontStyle.Italic))));
      this.eStatus.Location = new System.Drawing.Point(0, 248);
      this.eStatus.ReadOnly = true;
      this.eStatus.Size = new System.Drawing.Size(240, 19);
      // 
      // lbFingerprintKO
      // 
      this.lbFingerprintKO.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Bold);
      this.lbFingerprintKO.ForeColor = System.Drawing.Color.OrangeRed;
      this.lbFingerprintKO.Location = new System.Drawing.Point(144, 80);
      this.lbFingerprintKO.Size = new System.Drawing.Size(88, 16);
      this.lbFingerprintKO.Text = "Fingerprint KO";
      this.lbFingerprintKO.Visible = false;
      // 
      // lbFingerprintOK
      // 
      this.lbFingerprintOK.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Bold);
      this.lbFingerprintOK.ForeColor = System.Drawing.Color.Green;
      this.lbFingerprintOK.Location = new System.Drawing.Point(144, 80);
      this.lbFingerprintOK.Size = new System.Drawing.Size(96, 16);
      this.lbFingerprintOK.Text = "Fingerprint OK";
      this.lbFingerprintOK.Visible = false;
      // 
      // btnLoad
      // 
      this.btnLoad.Enabled = false;
      this.btnLoad.Location = new System.Drawing.Point(92, 224);
      this.btnLoad.Size = new System.Drawing.Size(56, 20);
      this.btnLoad.Text = "Load";
      this.btnLoad.Click += new System.EventHandler(this.btnLoad_Click);
      // 
      // btnSave
      // 
      this.btnSave.Enabled = false;
      this.btnSave.Location = new System.Drawing.Point(92, 200);
      this.btnSave.Size = new System.Drawing.Size(56, 20);
      this.btnSave.Text = "Save";
      this.btnSave.Click += new System.EventHandler(this.btnSave_Click);
      // 
      // openFileDialog
      // 
      this.openFileDialog.Filter = "Finger TK files|*.fingertk";
      // 
      // saveFileDialog
      // 
      this.saveFileDialog.Filter = "Finger TK files|*.fingertk";
      // 
      // lbCard
      // 
      this.lbCard.Location = new System.Drawing.Point(0, 184);
      this.lbCard.Size = new System.Drawing.Size(48, 20);
      this.lbCard.Text = "Card :";
      // 
      // lbFinger
      // 
      this.lbFinger.Location = new System.Drawing.Point(160, 184);
      this.lbFinger.Size = new System.Drawing.Size(92, 20);
      this.lbFinger.Text = "Fingerprints :";
      // 
      // lbFile
      // 
      this.lbFile.Location = new System.Drawing.Point(92, 184);
      this.lbFile.Size = new System.Drawing.Size(72, 20);
      this.lbFile.Text = "File :";
      // 
      // fMain
      // 
      this.ClientSize = new System.Drawing.Size(242, 271);
      this.Controls.Add(this.pImages);
      this.Controls.Add(this.eLastName);
      this.Controls.Add(this.eFirstName);
      this.Controls.Add(this.eUserID);
      this.Controls.Add(this.btnWrite);
      this.Controls.Add(this.btnRead);
      this.Controls.Add(this.btnSave);
      this.Controls.Add(this.btnLoad);
      this.Controls.Add(this.btnEnroll);
      this.Controls.Add(this.btnVerify);
      this.Controls.Add(this.eStatus);
      this.Controls.Add(this.lbFinger);
      this.Controls.Add(this.lbFile);
      this.Controls.Add(this.lbFingerprintOK);
      this.Controls.Add(this.lbFingerprintKO);
      this.Controls.Add(this.lbAppId);
      this.Controls.Add(this.lbWeb);
      this.Controls.Add(this.lbDesc);
      this.Controls.Add(this.lbName);
      this.Controls.Add(this.lbUserID);
      this.Controls.Add(this.lbCard);
      this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedDialog;
      this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
      this.MaximizeBox = false;
      this.MinimizeBox = false;
      this.Text = "Fingerprints TK";
      this.Closed += new System.EventHandler(this.fMain_Closed);
      this.Closing += new System.ComponentModel.CancelEventHandler(this.fMain_Closing);

    }
#endregion

#region Windows Form Glue

    public fMain()
    {
      InitializeComponent();

      workerAllowed = true;
      workerRunning = false;
      workerNextTask = 0;
      workerDoneTask = 0;
      workerAfterTask = 0;
      workerResult = false;
      readerActive = false;

      workerEvent = new ManualResetEvent(false);
      workerThread = new Thread(new ThreadStart(workerLoop));
      workerThread.Start();

      workerStart(WORK_OPEN_READER);
    }

    private void btnVerify_Click(object sender, System.EventArgs e)
    {
      UpdateStatus(false);
      workerStart(WORK_VERIFY);
    }

    private void btnEnroll_Click(object sender, System.EventArgs e)
    {
      UpdateStatus(false);
      workerStart(WORK_ENROLL);
    }

    private void btnRead_Click(object sender, System.EventArgs e)
    {
      eUserID.Text = "";
      eLastName.Text = "";
      eFirstName.Text = "";

      UpdateStatus(false, false, false);

      workerAfterTask = WORK_READ_CARD;
      workerStart(WORK_FIND_CARD);
    }

    private void btnWrite_Click(object sender, System.EventArgs e)
    {
      UpdateStatus(false);

      /* Retrieve updated data */
      while (eUserID.Text.Length < 16)
        eUserID.Text = "0" + eUserID.Text;
      fromHEX(card_user_data, 0, 8, eUserID.Text);
      fromSTR(card_user_data, 16, 32, eLastName.Text);
      fromSTR(card_user_data, 48, 32, eFirstName.Text);
      card_user_data[10] = card_finger_size_1;
      card_user_data[11] = card_finger_size_2;
      card_user_data[12] = card_finger_type;

      workerAfterTask = WORK_WRITE_CARD;
      workerStart(WORK_FIND_CARD);
    }

    private void fMain_Closed(object sender, System.EventArgs e)
    {
      if (readerActive)
        SPROX.ReaderClose();
    }


    protected override void Dispose(bool disposing)
    {
      base.Dispose(disposing);
    }

    private void show_enroll_image(System.Windows.Forms.PictureBox image)
    {
      pImages.Visible = true;
      imgEnrollLeft1.Visible = false;
      imgEnrollLeft2.Visible = false;
      imgEnrollLeft3.Visible = false;
      imgEnrollRight1.Visible = false;
      imgEnrollRight2.Visible = false;
      imgEnrollRight3.Visible = false;
      imgVerify.Visible = false;
      if (image != null)
        image.Visible = true;
      show_async_image(imgPlaceFinger);
    }

    private void show_async_image(System.Windows.Forms.PictureBox image)
    {
      pImages.Visible = true;
      imgFingerOK.Visible = false;
      imgMoveRight.Visible = false;
      imgMoveLeft.Visible = false;
      imgMoveDown.Visible = false;
      imgMoveUp.Visible = false;
      imgPlaceFinger.Visible = false;
      imgPressHarder.Visible = false;
      if (image != null)
        image.Visible = true;
    }

    private void fMain_Closing(object sender,
                               System.ComponentModel.CancelEventArgs e)
    {
      int retry = 50;

      workerAllowed = false;
      workerEvent.Set();

      while ((workerRunning) && (retry-- > 0))
        Thread.Sleep(20);
    }

    private void btnSave_Click(object sender, System.EventArgs e)
    {
      if (saveFileDialog.InitialDirectory == "")
        saveFileDialog.InitialDirectory =
          CodePPC.General.Tools.GetApplicationPath();

      saveFileDialog.FileName = "User_" + eUserID.Text;
      if (saveFileDialog.ShowDialog() != DialogResult.OK)
        return;

      eStatus.Text = "Saving user's fingerprint and data";

      StreamWriter wr = File.CreateText(saveFileDialog.FileName);

      if (wr == null)
      {
        eStatus.Text = "Failed to save the data";
        return;
      }

      wr.WriteLine("id={0}", eUserID.Text);
      wr.WriteLine("lname={0}", eLastName.Text);
      wr.WriteLine("fname={0}", eFirstName.Text);

      string s;

      s = string.Format("{0:X02}", card_finger_type);
      s += ",";
      s += toHEX(card_finger_data_1, 0, card_finger_size_1);
      wr.WriteLine("finger1={0}", s);

      s = string.Format("{0:X02}", card_finger_type);
      s += ",";
      s += toHEX(card_finger_data_2, 0, card_finger_size_2);
      wr.WriteLine("finger2={0}", s);

      wr.Close();

      eStatus.Text = "Data have been saved";
    }

    private void btnLoad_Click(object sender, System.EventArgs e)
    {
      if (openFileDialog.InitialDirectory == "")
        openFileDialog.InitialDirectory =
          CodePPC.General.Tools.GetApplicationPath();
      if (openFileDialog.ShowDialog() != DialogResult.OK)
        return;

      eStatus.Text = "Loading data file";

      StreamReader sr = File.OpenText(openFileDialog.FileName);

      if (sr == null)
      {
        UpdateStatus(true, false, false);
        eStatus.Text = "Failed to load the data";
        return;
      }

      for (int i = 0; i < 80; i++)
        card_user_data[i] = 0;

      eUserID.Text = "";
      eLastName.Text = "";
      eFirstName.Text = "";
      card_finger_type = 0;
      card_finger_size_1 = 0;
      card_finger_size_2 = 0;

      while (sr.Peek() >= 0)
      {
        string l;

        string[]s;

        l = sr.ReadLine();
        s = l.Split('=');
        if (s.Length != 2)
          continue;

        switch (s[0])
        {
          case "id":
            eUserID.Text = s[1];
            break;
          case "lname":
            eLastName.Text = s[1];
            break;
          case "fname":
            eFirstName.Text = s[1];
            break;
          case "finger1":
            s = s[1].Split(',');
            if (s.Length != 2)
              break;
            card_finger_type = Convert.ToByte(s[0], 16);
            card_finger_size_1 = (byte) (s[1].Length / 2);
            fromHEX(card_finger_data_1, 0, 170, s[1]);
            break;
          case "finger2":
            s = s[1].Split(',');
            if (s.Length != 2)
              break;
            card_finger_type = Convert.ToByte(s[0], 16);
            card_finger_size_2 = (byte) (s[1].Length / 2);
            fromHEX(card_finger_data_2, 0, 170, s[1]);
            break;
        }
      }

      sr.Close();

      while (eUserID.Text.Length < 16)
        eUserID.Text = "0" + eUserID.Text;
      fromHEX(card_user_data, 0, 8, eUserID.Text);
      fromSTR(card_user_data, 16, 32, eLastName.Text);
      fromSTR(card_user_data, 48, 32, eFirstName.Text);

      UpdateStatus(true, true, false);
      eStatus.Text = "Data have been loaded";
    }

    private bool was_loaded, was_editable;

    public void UpdateStatus(bool active)
    {
      UpdateStatus(active, was_loaded, was_editable);
    }

    public void UpdateStatus(bool active, bool loaded, bool editable)
    {
      lbFingerprintOK.Visible = false;
      lbFingerprintKO.Visible = false;

      btnVerify.Enabled = (active) && (loaded) && ((card_finger_size_1 > 0)
                                                   || (card_finger_size_2 >
                                                       0));
      btnEnroll.Enabled = (active);
      btnLoad.Enabled = (active);
      btnSave.Enabled = (active) && (loaded);
      btnWrite.Enabled = (active) && (loaded);
      btnRead.Enabled = (active);

      if (editable)
      {
        eUserID.ReadOnly = false;
        eLastName.ReadOnly = false;
        eFirstName.ReadOnly = false;
        eUserID.BackColor = System.Drawing.SystemColors.Window;
        eLastName.BackColor = System.Drawing.SystemColors.Window;
        eFirstName.BackColor = System.Drawing.SystemColors.Window;
      } else
      {
        eUserID.ReadOnly = true;
        eLastName.ReadOnly = true;
        eFirstName.ReadOnly = true;
        eUserID.BackColor = System.Drawing.SystemColors.Control;
        eLastName.BackColor = System.Drawing.SystemColors.Control;
        eFirstName.BackColor = System.Drawing.SystemColors.Control;
      }

      was_loaded = loaded;
      was_editable = editable;
    }

    static void Main()
    {
      Application.Run(new fMain());
    }

#endregion

  }
}
