#include "stdafx.h"
#include "finger_tk.h"
#include "finger_tkDlg.h"
#include "finger_tkThread.h"
#include "common.h"

#include "hardware/springprox/springprox.h"
#include "sagem_mso_ilv.h"

static const unsigned char MIFARE_HEADER[32] =
  { 0x00, 0x00, 'P', 'r', 'o', '-', 'A', 'c', 't', 'i', 'v', 'e', 0x00, 0x11,
  0x22, 0x33
};
static const unsigned char MIFARE_KEY_A[6] =
  { 0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5 };
static const unsigned char MIFARE_KEY_B[6] =
  { 0x50, 0x41, 0x00, 0x11, 0x22, 0x33 };

/*
 * Read / Write Pro-Active formated tags
 * -------------------------------------
 */
BOOL read_from_tag_mifare(FINGERTK_DATA_ST * data)
{
  unsigned char buffer[11 * 48];
  unsigned char sector;
  short rc;

  dlg.display(_T("Reading Mifare card..."), NULL);

  for (sector = 0; sector < 11; sector++)
  {
    rc =
      SPROX_MifStReadSector(NULL, sector, &buffer[48 * sector], MIFARE_KEY_A);
    if (rc != MI_OK)
    {
      dlg.display(NULL, _T("Read error"));
      return false;
    }

    if (sector == 0)
    {
      /* Check that the expected header is found on blocks 1 & 2 */
      if (memcmp(&buffer[16], MIFARE_HEADER, 16))
      {
        dlg.display(NULL, _T("Card doesn't have our header"));
        return false;
      }
    }
  }

  /* Translate information block into interesting content... */
  /* Information block starts at offset 48 */
  memcpy(&data->user, &buffer[48], sizeof(data->user));

  /* Read the fingerprint file (if some) */
  /* Fingerprint block starts at offset 128 */
  if (data->user.finger_size[0] || data->user.finger_size[1])
  {
    memcpy(data->fingers, &buffer[128], sizeof(data->fingers));
  }

  return true;
}

BOOL read_from_tag_sagem(FINGERTK_DATA_ST * data)
{
  BYTE sect;
  BYTE buffer[13][48];          /* 13 sectors of 48 bytes each = 648bytes */
  SWORD rc;

  dlg.display(_T("Reading Mifare tag..."), NULL);
  for (sect = 1; sect <= 13; sect++)
  {
    /* We read with pre-loaded keys (RAM or EEPROM) */
    rc = SPROX_MifStReadSector(NULL, sect, buffer[sect - 1], NULL);
    if (rc != MI_OK)
    {
      dlg.display(NULL, _T("Failed to read the tag"));
      return FALSE;
    }
  }

  {
    /* Try to find a Sagem formated string */
    BYTE tag, idx;
    WORD length, offset;
    BYTE *value;

    offset = 0;
    idx = 0;

    while ((offset < sizeof(buffer))
           && ilv_loop(buffer[0], &offset, &tag, &length, &value))
    {
      if ((tag == 0x00) || (tag == 0xFF))
        break;

      switch (tag)
      {
        case 0x32:             /* User ID */
          fromHEXa(data->user.user_id, sizeof(data->user.user_id),
                   (char *) value);
          break;

        case 0x20:             /* User Name */
          if (length > sizeof(data->user.last_name))
            length = sizeof(data->user.last_name);
          memcpy(data->user.last_name, value, length);
          break;

        case 0x30:             /* Finger 1 */
          if (length <= sizeof(data->fingers) / 2)
          {
            data->user.finger_type = ID_PKCOMP; /* Assume it is a PKCOMP minutiae */
            data->user.finger_size[0] = (BYTE) length;
            memcpy(data->fingers[0], value, length);
          }
          break;

        case 0x31:             /* Finger 2 */
          if (length <= sizeof(data->fingers) / 2)
          {
            data->user.finger_type = ID_PKCOMP; /* Assume it is a PKCOMP minutiae */
            data->user.finger_size[1] = (BYTE) length;
            memcpy(data->fingers[1], value, length);
          }
          break;

      }
    }
  }

  if ((data->user.finger_size[0] == 0) && (data->user.finger_size[0] == 0))
  {
    dlg.display(NULL, _T("Tag is not formated"));
    return FALSE;
  }

  dlg.display(_T("Mifare tag read"), _T(""));
  return TRUE;
}

BOOL write_to_tag_mifare(FINGERTK_DATA_ST * data)
{
  BYTE buffer[528];
  int i;
  BYTE sector;
  short rc;

  dlg.display(_T("Writing Mifare tag..."), _T(""));

  memset(buffer, 0, sizeof(buffer));

  /* Build the header */
  for (i = 0; i < 16; i++)
    buffer[16 + i] = MIFARE_HEADER[i];

  /* Retrieve the information block */
  memcpy(&buffer[48], &data->user, sizeof(data->user));

  /* Retrieve fingerprints */
  if (data->user.finger_size[0] || data->user.finger_size[1])
  {
    memcpy(&buffer[128], data->fingers, sizeof(data->fingers));
  }

  /* Write the part common to 1k and 4k */
  for (sector = 0; sector < 11; sector++)
  {
    /* Try to write with the expected key */
    rc =
      SPROX_MifStWriteSector(NULL, sector, &buffer[48 * sector],
                             MIFARE_KEY_B);
    if (rc != MI_OK)
    {
      /* Try to format the sector */
      rc =
        SPROX_MifStUpdateAccessBlock(NULL, sector, NULL, MIFARE_KEY_A,
                                     MIFARE_KEY_B, ACC_BLOCK_VALUE,
                                     ACC_BLOCK_VALUE, ACC_BLOCK_VALUE,
                                     ACC_AUTH_NORMAL);
      if (rc != MI_OK)
      {
        dlg.display(NULL, _T("Unable to format the Mifare tag"));
        return false;
      }
      /* Retry to write with the expected key */
      rc =
        SPROX_MifStWriteSector(NULL, sector, &buffer[48 * sector],
                               MIFARE_KEY_B);
      if (rc != MI_OK)
      {
        dlg.display(NULL, _T("Failed to write to the Mifare tag"));
        return false;
      }
    }
  }

  dlg.display(_T("Mifare tag written"), _T(""));
  return true;
}
